@@ -40,6 +40,7 @@
 #include <camel/camel-data-wrapper.h>
 #include <camel/camel-multipart.h>
 #include <camel/camel-private.h>
+#include <camel/camel-session.h>
 #include <camel/camel-stream-buffer.h>
 #include <camel/camel-stream-mem.h>
 #include <camel/camel-debug.h>
@@ -54,6 +55,7 @@
 #include "camel-mapi-folder.h"
 #include "camel-mapi-private.h"
 #include "camel-mapi-summary.h"
+#include "camel-mapi-utils.h"
 
 #define DEBUG_FN( ) printf("----%u %s\n", (unsigned int)pthread_self(), __FUNCTION__);
 #define SUMMARY_FETCH_BATCH_COUNT 150
@@ -183,16 +185,100 @@ mapi_item_free (MapiItem *item, gpointer data)
 {
 	g_free (item->header.subject);
 	g_free (item->header.from);
-	/* g_free (item->header.from_email); */
-	/* g_free (item->header.from_type); */
+
 	g_free (item->header.to);
 	g_free (item->header.cc);
 	g_free (item->header.bcc);
 
+	g_free (item->header.references);
+	g_free (item->header.message_id);
+	g_free (item->header.in_reply_to);
+
 	exchange_mapi_util_free_attachment_list (&item->attachments);
 	exchange_mapi_util_free_stream_list (&item->generic_streams);
 }
 
+static gboolean
+read_item_common (MapiItem *item, uint32_t ulPropTag, gconstpointer prop_data)
+{
+	gboolean found = TRUE;
+
+	#define sv(_x,_y) G_STMT_START { g_free (_x); _x = _y; } G_STMT_END
+
+	switch (ulPropTag) {
+	case PR_INTERNET_CPID: {
+		const uint32_t *ui32 = (const uint32_t *) prop_data;
+		if (ui32)
+			item->header.cpid = *ui32;
+		} break;
+	/* FIXME : Instead of duping. Use talloc_steal to reuse the memory */
+	case PR_SUBJECT:
+		sv (item->header.subject, utf8tolinux (prop_data));
+		break;
+	case PR_SUBJECT_UNICODE :
+		sv (item->header.subject, g_strdup (prop_data));
+		break;
+	case PR_DISPLAY_TO :
+		sv (item->header.to, utf8tolinux (prop_data));
+		break;
+	case PR_DISPLAY_TO_UNICODE :
+		sv (item->header.to, g_strdup (prop_data));
+		break;
+	case PR_DISPLAY_CC:
+		sv (item->header.cc, utf8tolinux (prop_data));
+		break;
+	case PR_DISPLAY_CC_UNICODE:
+		sv (item->header.cc, g_strdup (prop_data));
+		break;
+	case PR_DISPLAY_BCC:
+		sv (item->header.bcc, utf8tolinux (prop_data));
+		break;
+	case PR_DISPLAY_BCC_UNICODE:
+		sv (item->header.bcc, g_strdup (prop_data));
+		break;
+	case PR_SENT_REPRESENTING_NAME:
+		sv (item->header.from, utf8tolinux (prop_data));
+		break;
+	case PR_SENT_REPRESENTING_NAME_UNICODE:
+		sv (item->header.from, g_strdup (prop_data));
+		break;
+	case PR_SENT_REPRESENTING_EMAIL_ADDRESS:
+		sv (item->header.from_email, utf8tolinux (prop_data));
+		break;
+	case PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE:
+		sv (item->header.from_email, g_strdup (prop_data));
+		break;
+	case PR_SENT_REPRESENTING_ADDRTYPE:
+		sv (item->header.from_type, utf8tolinux (prop_data));
+		break;
+	case PR_SENT_REPRESENTING_ADDRTYPE_UNICODE:
+		sv (item->header.from_type, g_strdup (prop_data));
+		break;
+	case PR_MESSAGE_SIZE:
+		item->header.size = *(glong *)prop_data;
+		break;
+	case PR_INTERNET_MESSAGE_ID:
+		item->header.message_id = g_strdup (prop_data);
+		break;
+	case PR_INTERNET_REFERENCES:
+		item->header.references = g_strdup (prop_data);
+		break;
+	case PR_IN_REPLY_TO_ID:
+		item->header.in_reply_to = g_strdup (prop_data);
+		break;
+	case PR_TRANSPORT_MESSAGE_HEADERS:
+		item->header.transport_headers = utf8tolinux (prop_data);
+		break;
+	default:
+		found = FALSE;
+		break;
+	}
+
+	#undef sv
+
+	return found;
+}
+
 static gboolean
 fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 {
@@ -207,7 +293,6 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 	struct timeval fi_data_mod_time = { 0 };
 	guint32 j = 0;
 	NTTIME ntdate;
-	gchar *from_name, *from_email;
 
 	MapiItem *item = g_new0(MapiItem , 1);
 
@@ -223,42 +308,12 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 	item->recipients = item_data->recipients;
 
 	for (j = 0; j < item_data->properties->cValues; j++) {
-
 		gconstpointer prop_data = get_mapi_SPropValue_data(&item_data->properties->lpProps[j]);
 
+		if (read_item_common (item, item_data->properties->lpProps[j].ulPropTag, prop_data))
+			continue;
+
 		switch (item_data->properties->lpProps[j].ulPropTag) {
-		/* FIXME : Instead of duping. Use talloc_steal to reuse the memory */
-		case PR_NORMALIZED_SUBJECT:
-		case PR_NORMALIZED_SUBJECT_UNICODE :
-			item->header.subject = g_strdup (prop_data);
-			break;
-		case PR_DISPLAY_TO :
-		case PR_DISPLAY_TO_UNICODE :
-			item->header.to = g_strdup (prop_data);
-			break;
-		case PR_DISPLAY_CC:
-		case PR_DISPLAY_CC_UNICODE:
-			item->header.cc = g_strdup (prop_data);
-			break;
-		case PR_DISPLAY_BCC:
-		case PR_DISPLAY_BCC_UNICODE:
-			item->header.bcc = g_strdup (prop_data);
-			break;
-		case PR_SENT_REPRESENTING_NAME:
-		case PR_SENT_REPRESENTING_NAME_UNICODE:
-			item->header.from = g_strdup (prop_data);
-			break;
-		case PR_SENT_REPRESENTING_EMAIL_ADDRESS:
-		case PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE:
-			item->header.from_email = g_strdup (prop_data);
-			break;
-		case PR_SENT_REPRESENTING_ADDRTYPE:
-		case PR_SENT_REPRESENTING_ADDRTYPE_UNICODE:
-			item->header.from_type = g_strdup (prop_data);
-			break;
-		case PR_MESSAGE_SIZE:
-			item->header.size = *(glong *)prop_data;
-			break;
 		case PR_MESSAGE_DELIVERY_TIME:
 			delivery_date = (struct FILETIME *) prop_data;
 			break;
@@ -321,6 +376,80 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 	return TRUE;
 }
 
+static void
+mapi_set_message_id (CamelMapiMessageInfo *mapi_mi, const gchar *message_id)
+{
+	gchar *msgid;
+	guint8 *digest;
+	gsize length;
+	CamelMessageInfoBase *mi = &mapi_mi->info;
+
+	msgid = camel_header_msgid_decode (message_id);
+	if (msgid) {
+		GChecksum *checksum;
+
+		length = g_checksum_type_get_length (G_CHECKSUM_MD5);
+		digest = g_alloca (length);
+
+		checksum = g_checksum_new (G_CHECKSUM_MD5);
+		g_checksum_update (checksum, (guchar *) msgid, -1);
+		g_checksum_get_digest (checksum, digest, &length);
+		g_checksum_free (checksum);
+
+		memcpy(mi->message_id.id.hash, digest, sizeof(mi->message_id.id.hash));
+		g_free(msgid);
+	}
+
+}
+
+static void
+mapi_set_message_references (CamelMapiMessageInfo *mapi_mi, const gchar *references, const gchar *in_reply_to)
+{
+	struct _camel_header_references *refs, *irt, *scan;
+	guint8 *digest;
+	gint count;
+	gsize length;
+	CamelMessageInfoBase *mi = &mapi_mi->info;
+
+	refs = camel_header_references_decode (references);
+	irt = camel_header_references_inreplyto_decode (in_reply_to);
+	if (refs || irt) {
+		if (irt) {
+			/* The References field is populated from the "References" and/or "In-Reply-To"
+			   headers. If both headers exist, take the first thing in the In-Reply-To header
+			   that looks like a Message-ID, and append it to the References header. */
+
+			if (refs)
+				irt->next = refs;
+
+			refs = irt;
+		}
+
+		count = camel_header_references_list_size(&refs);
+		mi->references = g_malloc(sizeof(*mi->references) + ((count-1) * sizeof(mi->references->references[0])));
+
+		length = g_checksum_type_get_length (G_CHECKSUM_MD5);
+		digest = g_alloca (length);
+
+		count = 0;
+		scan = refs;
+		while (scan) {
+			GChecksum *checksum;
+
+			checksum = g_checksum_new (G_CHECKSUM_MD5);
+			g_checksum_update (checksum, (guchar *) scan->id, -1);
+			g_checksum_get_digest (checksum, digest, &length);
+			g_checksum_free (checksum);
+
+			memcpy(mi->references->references[count].id.hash, digest, sizeof(mi->message_id.id.hash));
+			count++;
+			scan = scan->next;
+		}
+		mi->references->size = count;
+		camel_header_references_list_clear(&refs);
+	}
+}
+
 static void
 mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **changeinfo,
 		   CamelException *ex, gboolean uid_flag) 
@@ -333,7 +462,7 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 	CamelFolderChangeInfo *changes = NULL;
 	gboolean exists = FALSE;
 	GString *str = g_string_new (NULL);
-	const gchar *folder_id = NULL;
+	const gchar *from_email, *folder_id = NULL;
 	GSList *item_list = list;
 	int total_items = g_slist_length (item_list), i=0;
 
@@ -346,12 +475,12 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 		return;
 	}
 
-	camel_operation_start (NULL, _("Fetching summary information for new messages in %s"), folder->name);
+	camel_operation_start (NULL, _("Updating local summary cache for new messages in %s"), folder->name);
 
 	for ( ; item_list != NULL ; item_list = g_slist_next (item_list) ) {
 		MapiItem *temp_item ;
 		MapiItem *item;
-		gchar *msg_uid, *to = NULL, *from = NULL;
+		gchar *msg_uid, *to = NULL, *from = NULL, *cc = NULL;
 		guint64 id;
 
 		exists = FALSE;
@@ -389,17 +518,23 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
 
 		if (!exists) {
  			GSList *l = NULL;
- 			guint32 i =0;
+ 			guint32 count_to = 0, count_cc =0;
 
 			mi->info.uid = exchange_mapi_util_mapi_ids_to_uid(item->fid, item->mid);
 			mi->info.subject = camel_pstring_strdup(item->header.subject);
 			mi->info.date_sent = mi->info.date_received = item->header.recieved_time;
 			mi->info.size = (guint32) item->header.size;
 
+			/*Threading related properties*/
+			mapi_set_message_id (mi, item->header.message_id);
+			if (item->header.references || item->header.in_reply_to)
+				mapi_set_message_references (mi, item->header.references, item->header.in_reply_to);
+
+			/*Recipients*/
  			for (l = item->recipients; l; l=l->next) {
- 				char *display_name, *name = NULL, *formatted_id;
- 				uint32_t rcpt_type = MAPI_TO;
- 				uint32_t *type = NULL; 
+ 				gchar *formatted_id = NULL;
+				const char *name, *display_name;
+ 				guint32 *type = NULL;
  				struct SRow aRow;
  				ExchangeMAPIRecipient *recip = (ExchangeMAPIRecipient *)(l->data);
  				
@@ -411,55 +546,77 @@ mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **ch
  				aRow.ulAdrEntryPad = 0;
  				aRow.cValues = recip->out.all_cValues;
  				aRow.lpProps = recip->out.all_lpProps;
- 				
+
  				type = (uint32_t *) find_SPropValue_data(&aRow, PR_RECIPIENT_TYPE);
- 
- 				if (*type == MAPI_TO) {
- 					/*Name is probably available in one of these props.*/
+
+				if (type) {
  					name = (const char *) find_SPropValue_data(&aRow, PR_DISPLAY_NAME);
  					name = name ? name : (const char *) find_SPropValue_data(&aRow, PR_RECIPIENT_DISPLAY_NAME);
  					name = name ? name : (const char *) find_SPropValue_data(&aRow, 
  												 PR_RECIPIENT_DISPLAY_NAME_UNICODE);
  					name = name ? name : (const char *) find_SPropValue_data(&aRow, 
  												 PR_7BIT_DISPLAY_NAME_UNICODE);
- 					display_name = name ? name : g_strdup (recip->email_id);
+ 					display_name = name ? name : recip->email_id;
  					formatted_id = camel_internet_address_format_address(display_name, recip->email_id);
- 
- 					/* hmm */
- 					if (i)
- 						to = g_string_append (to, ", ");
- 
- 					g_string_append_printf (to, "%s", formatted_id);
- 					i++;
- 				}
- 
- 				/*TODO : from ? */
+
+					switch (*type) {
+					case MAPI_TO:
+						if (count_to) {
+							to = g_strconcat (to, ", ", formatted_id, NULL);
+							g_free (formatted_id);
+						} else
+							to = formatted_id;
+						count_to ++;
+						break;
+
+					case MAPI_CC:
+						if (count_cc) {
+							cc = g_strconcat (cc, ", ", formatted_id, NULL);
+							g_free (formatted_id);
+						} else
+							cc = formatted_id;
+						count_cc ++;
+						break;
+
+					default:
+						continue;
+					}
+				}
  			}
  			
  			if ((item->header.from_type != NULL) && !g_utf8_collate (item->header.from_type, "EX")) {
  				CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
- 				item->header.from_email = exchange_mapi_util_ex_to_smtp (item->header.from_email);
+				from_email = exchange_mapi_util_ex_to_smtp (item->header.from_email);
  				CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+
+ 				g_free (item->header.from_email);
+				item->header.from_email = g_strdup (from_email);				
  			}
  
  			item->header.from_email = item->header.from_email ? 
  				item->header.from_email : item->header.from;
- 
- 			from = camel_internet_address_format_address (item->header.from, 
- 								      item->header.from_email);
- 			mi->info.from = camel_pstring_strdup (from);
- 			mi->info.to = camel_pstring_strdup (to);
 
+			if (item->header.from_email) {
+				from = camel_internet_address_format_address (item->header.from, 
+ 									      item->header.from_email);
+ 				mi->info.from = camel_pstring_strdup (from);
+			} else
+				mi->info.from = NULL;
 
+			/* Fallback */
+ 			mi->info.to = to ? camel_pstring_strdup (to) : g_strdup (item->header.to);
+ 			mi->info.cc = to ? camel_pstring_strdup (cc) : g_strdup (item->header.cc);
 		}
 
 		if (exists) {
 			camel_folder_change_info_change_uid (changes, mi->info.uid);
 			camel_message_info_free (pmi);
 		} else {
+			CAMEL_MAPI_FOLDER_REC_LOCK (folder, cache_lock);
 			camel_folder_summary_add (folder->summary,(CamelMessageInfo *)mi);
 			camel_folder_change_info_add_uid (changes, mi->info.uid);
 			camel_folder_change_info_recent_uid (changes, mi->info.uid);
+			CAMEL_MAPI_FOLDER_REC_UNLOCK (folder, cache_lock);
 		}
 
 		/********************* Summary ends *************************/
@@ -490,6 +647,137 @@ mapi_utils_do_flags_diff (flags_diff_t *diff, guint32 old, guint32 _new)
 	diff->bits = _new & diff->changed;
 }
 
+
+struct mapi_update_deleted_msg {
+	CamelSessionThreadMsg msg;
+
+	CamelFolder *folder;
+	mapi_id_t folder_id;
+};
+
+static gboolean
+deleted_items_sync_cb (FetchItemsCallbackData *item_data, gpointer data)
+{
+	GSList **uid_list = (GSList **) data;
+	gchar *msg_uid = exchange_mapi_util_mapi_ids_to_uid (item_data->fid, 
+							     item_data->mid);
+
+	*uid_list = g_slist_prepend (*uid_list, msg_uid);
+
+	/* Progress update */
+	if (item_data->total > 0)
+		camel_operation_progress (NULL, (item_data->index * 100)/item_data->total);
+
+	/* Check if we have to stop */
+	if (camel_operation_cancel_check(NULL))
+		return FALSE;
+
+	return TRUE;
+}
+
+static void
+mapi_sync_deleted (CamelSession *session, CamelSessionThreadMsg *msg)
+{
+	struct mapi_update_deleted_msg *m = (struct mapi_update_deleted_msg *)msg;
+
+	CamelMapiStore *mapi_store = CAMEL_MAPI_STORE (m->folder->parent_store);
+	CamelMapiFolder *mapi_folder = CAMEL_MAPI_FOLDER (m->folder);
+	CamelFolderChangeInfo *changes = NULL;
+	CamelMessageInfo *info = NULL;
+
+	guint32 index, count, options = 0;
+	GSList *server_uid_list = NULL;
+	const gchar *uid = NULL;
+	GSList *tmp_list_item = NULL;
+
+	/* Currently we don't have simple wrapper over getprops.*/
+	const guint32 prop_list[] = { PR_LAST_MODIFICATION_TIME };
+	
+	if (((CamelOfflineStore *) mapi_store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL ||
+			((CamelService *)mapi_store)->status == CAMEL_SERVICE_DISCONNECTED) {
+	
+		return;
+	}
+
+	camel_operation_start (NULL, _("Retrieving message IDs from server for %s"), m->folder->name);
+
+	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
+
+	/*Get the UID list from server.*/
+	exchange_mapi_connection_fetch_items  (m->folder_id, NULL, NULL,
+					       prop_list, G_N_ELEMENTS (prop_list),
+					       NULL, NULL,
+					       deleted_items_sync_cb, &server_uid_list,
+					       options);
+
+	camel_operation_end (NULL);
+
+	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+
+	/* Check if we have to stop */
+	if (camel_operation_cancel_check(NULL))
+		return;
+
+	changes = camel_folder_change_info_new ();
+
+	count = camel_folder_summary_count (m->folder->summary);
+	camel_operation_start (NULL, _("Removing deleted messages from cache in %s"), m->folder->name);
+
+	/* Iterate over cache and check if the UID is in server*/
+	for (index = 0; index < count; index++) {
+		/*FIXME :Any other list available ???*/
+		info = camel_folder_summary_index (m->folder->summary, index);
+		if (!info) continue; /*This is bad. *Should* not happen*/
+
+		uid = camel_message_info_uid (info);
+
+		if (server_uid_list) {
+			/* TODO : Find a better way and avoid this linear search */
+			tmp_list_item = g_slist_find_custom (server_uid_list, (gconstpointer) uid, 
+						  (GCompareFunc) g_strcmp0);
+		}
+
+		/* If it is not in server list, clean our cache */
+		if (!tmp_list_item->data && uid) {
+			CAMEL_MAPI_FOLDER_REC_LOCK (m->folder, cache_lock);
+			camel_folder_summary_remove_uid (m->folder->summary, uid);
+			camel_data_cache_remove (mapi_folder->cache, "cache", uid, NULL);
+			camel_folder_change_info_remove_uid (changes, uid);
+			CAMEL_MAPI_FOLDER_REC_UNLOCK (m->folder, cache_lock);
+		}
+
+		/* Progress update */
+		camel_operation_progress (NULL, (index * 100)/count); /* ;-) */
+
+		/* Check if we have to stop */
+		if (camel_operation_cancel_check(NULL))
+			return;
+	}
+
+	camel_operation_end (NULL);
+
+	camel_object_trigger_event (m->folder, "folder_changed", changes);
+	camel_folder_change_info_free (changes);
+
+	/* Discard server uid list */
+	g_slist_foreach (server_uid_list, (GFunc) g_free, NULL);
+	g_slist_free (server_uid_list);
+}
+
+static void
+mapi_sync_deleted_free (CamelSession *session, CamelSessionThreadMsg *msg)
+{
+	struct mapi_update_deleted_msg *m = (struct mapi_update_deleted_msg *)msg;
+
+	camel_object_unref (m->folder);
+	/* camel_session_thread_msg_free (session, &m->msg); */
+}
+
+static CamelSessionThreadOps deleted_items_sync_ops = {
+	mapi_sync_deleted,
+	mapi_sync_deleted_free,
+};
+
 static void
 mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 {
@@ -502,7 +790,7 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 	flags_diff_t diff, unset_flags;
 	const char *folder_id;
 	mapi_id_t fid, deleted_items_fid;
-	int count, i;
+	gint count, i;
 	guint32 options =0;
 
 	GSList *deleted_items, *deleted_head;
@@ -565,17 +853,14 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 						deleted_head = NULL;
 						deleted_head = deleted_items = g_slist_prepend (deleted_items, mid);
 					}
-
-					CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
-
-					}
-				}
-				
-				if (diff.bits & CAMEL_MESSAGE_SEEN) {
-					read_items = g_slist_prepend (read_items, mid);
-				} else if (unset_flags.bits & CAMEL_MESSAGE_SEEN) {
-					unread_items = g_slist_prepend (unread_items, mid);
 				}
+			}
+
+			if (diff.bits & CAMEL_MESSAGE_SEEN) {
+				read_items = g_slist_prepend (read_items, mid);
+			} else if (unset_flags.bits & CAMEL_MESSAGE_SEEN) {
+				unread_items = g_slist_prepend (unread_items, mid);
+			}
 		}
 		camel_message_info_free (info);
 	}
@@ -595,6 +880,7 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 		g_slist_free (read_items);
 	}
 
+	/* Remove messages from server*/
 	if (deleted_items) {
 		CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 		if (mapi_folder->type & CAMEL_FOLDER_TYPE_TRASH) {
@@ -606,7 +892,8 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 
 		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 	}
-	/*Remove them from cache*/
+
+	/*Remove messages from local cache*/
 	while (deleted_items) {
 		char* deleted_msg_uid = g_strdup_printf ("%016" G_GINT64_MODIFIER "X%016" G_GINT64_MODIFIER "X", fid, *(mapi_id_t *)deleted_items->data);
 
@@ -618,18 +905,13 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 		deleted_items = g_slist_next (deleted_items);
 	}
 
-
 	if (unread_items) {
-		CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 		/* TODO */
-		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 		g_slist_free (unread_items);
 	}
 
 	if (expunge) {
-		CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 		/* TODO */
-		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 	}
 
 	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
@@ -645,18 +927,23 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 	CamelMapiStore *mapi_store = CAMEL_MAPI_STORE (folder->parent_store);
 	CamelMapiFolder *mapi_folder = CAMEL_MAPI_FOLDER (folder);
 	CamelMapiSummary *mapi_summary = CAMEL_MAPI_SUMMARY (folder->summary);
+	CamelSession *session = ((CamelService *)folder->parent_store)->session;
+
 	gboolean is_proxy = folder->parent_store->flags & CAMEL_STORE_PROXY;
 	gboolean is_locked = FALSE;
 	gboolean status;
 
 	struct mapi_SRestriction *res = NULL;
 	struct SSortOrderSet *sort = NULL;
+	struct mapi_update_deleted_msg *deleted_items_op_msg;
 	fetch_items_data *fetch_data = g_new0 (fetch_items_data, 1);
 
 	const gchar *folder_id = NULL;
 
 	const guint32 summary_prop_list[] = {
-		PR_NORMALIZED_SUBJECT,
+		PR_INTERNET_CPID,
+		PR_SUBJECT,
+		PR_SUBJECT_UNICODE,
 		PR_MESSAGE_SIZE,
 		PR_MESSAGE_DELIVERY_TIME,
 		PR_MESSAGE_FLAGS,
@@ -664,9 +951,13 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 		PR_SENT_REPRESENTING_EMAIL_ADDRESS,
 		PR_SENT_REPRESENTING_ADDRTYPE,
 		PR_LAST_MODIFICATION_TIME,
+		PR_INTERNET_MESSAGE_ID,
+		PR_INTERNET_REFERENCES,
+		PR_IN_REPLY_TO_ID,
 		PR_DISPLAY_TO,
 		PR_DISPLAY_CC,
-		PR_DISPLAY_BCC
+		PR_DISPLAY_BCC,
+		PR_TRANSPORT_MESSAGE_HEADERS
 	};
 
 	if (((CamelOfflineStore *) mapi_store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL)
@@ -736,9 +1027,11 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 			/*BUG : Fix exception string.*/
 			camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_UNAVAILABLE,
 					     _("This message is not available in offline mode."));
-			goto end2;
+			goto end1;
 		}
 
+		options |= MAPI_OPTIONS_FETCH_RECIPIENTS;
+
 		if (((CamelMapiFolder *)folder)->type & CAMEL_MAPI_FOLDER_PUBLIC)
 			options |= MAPI_OPTIONS_USE_PFSTORE;
 
@@ -753,7 +1046,7 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 
 		if (!status) {
 			camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_INVALID, _("Fetching items failed"));
-			goto end2;
+			goto end1;
 		}
 
 		/*Preserve last_modification_time from this fetch for later use with restrictions.*/
@@ -762,17 +1055,22 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 		camel_folder_summary_touch (folder->summary);
 		mapi_sync_summary (folder, ex);
 
-		camel_object_trigger_event (folder, "folder_changed", fetch_data->changes);
+		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+		is_locked = FALSE;
 
+		/* Downsync deleted items */
+		deleted_items_op_msg = camel_session_thread_msg_new (session, &deleted_items_sync_ops, 
+							 sizeof (*deleted_items_op_msg));
+		deleted_items_op_msg->folder = folder;
+		deleted_items_op_msg->folder_id = temp_folder_id;
+		camel_object_ref (folder);
+
+		camel_session_thread_queue (session, &deleted_items_op_msg->msg, 0);
+
+		camel_object_trigger_event (folder, "folder_changed", fetch_data->changes);
 		camel_folder_change_info_free (fetch_data->changes);
 	}
 
-
-	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
-	is_locked = FALSE;
-
-end2:
-	//TODO:
 end1:
 	if (is_locked)
 		CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
@@ -785,7 +1083,9 @@ end1:
 static const uint32_t camel_GetPropsList[] = {
 	PR_FID, 
 	PR_MID, 
+	PR_INTERNET_CPID,
 
+	PR_TRANSPORT_MESSAGE_HEADERS,
 	PR_MESSAGE_CLASS, 
 	PR_MESSAGE_CLASS_UNICODE, 
 	PR_MESSAGE_SIZE, 
@@ -795,11 +1095,14 @@ static const uint32_t camel_GetPropsList[] = {
 
 	PR_SUBJECT, 
 	PR_SUBJECT_UNICODE, 
-	PR_NORMALIZED_SUBJECT, 
-	PR_NORMALIZED_SUBJECT_UNICODE, 
 	PR_CONVERSATION_TOPIC, 
 	PR_CONVERSATION_TOPIC_UNICODE, 
 
+	/*Properties used for message threading.*/
+	PR_INTERNET_MESSAGE_ID,
+	PR_INTERNET_REFERENCES,
+	PR_IN_REPLY_TO_ID,
+
 	PR_BODY, 
 	PR_BODY_UNICODE, 
 	PR_HTML,
@@ -952,39 +1255,10 @@ fetch_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 
 		gconstpointer prop_data = get_mapi_SPropValue_data(&item_data->properties->lpProps[j]);
 
+		if (read_item_common (item, item_data->properties->lpProps[j].ulPropTag, prop_data))
+			continue;
+
 		switch (item_data->properties->lpProps[j].ulPropTag) {
-		/*FIXME : Instead of duping. Use talloc_steal to reuse the memory*/
-		case PR_NORMALIZED_SUBJECT:
-		case PR_NORMALIZED_SUBJECT_UNICODE :
-			item->header.subject = g_strdup (prop_data);
-			break;
-		case PR_DISPLAY_TO :
-		case PR_DISPLAY_TO_UNICODE :
-			item->header.to = g_strdup (prop_data);
-			break;
-		case PR_DISPLAY_CC:
-		case PR_DISPLAY_CC_UNICODE:
-			item->header.cc = g_strdup (prop_data);
-			break;
-		case PR_DISPLAY_BCC:
-		case PR_DISPLAY_BCC_UNICODE:
-			item->header.bcc = g_strdup (prop_data);
-			break;
-		case PR_SENT_REPRESENTING_NAME:
-		case PR_SENT_REPRESENTING_NAME_UNICODE:
-			item->header.from = g_strdup (prop_data);
-			break;
-		case PR_SENT_REPRESENTING_ADDRTYPE:
-		case PR_SENT_REPRESENTING_ADDRTYPE_UNICODE:
-			item->header.from_type = g_strdup (prop_data);
-			break;
-		case PR_SENT_REPRESENTING_EMAIL_ADDRESS:
-		case PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE:
-			item->header.from_email = g_strdup (prop_data);
-			break;
-		case PR_MESSAGE_SIZE:
-			item->header.size = *(glong *)prop_data;
-			break;
 		case PR_MESSAGE_CLASS:
 		case PR_MESSAGE_CLASS_UNICODE:
 			msg_class = (const char *) prop_data;
@@ -1008,7 +1282,7 @@ fetch_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 		body = g_new0(ExchangeMAPIStream, 1);
 		body->proptag = PR_BODY;
 		body->value = g_byte_array_new ();
-		body->value = g_byte_array_append (body->value, appointment_body_str, g_utf8_strlen ((const gchar *)appointment_body_str, -1));
+		body->value = g_byte_array_append (body->value, appointment_body_str, strlen ((const char *)appointment_body_str));
 
 		item->msg.body_parts = g_slist_append (item->msg.body_parts, body);
 
@@ -1046,17 +1320,18 @@ fetch_item_cb (FetchItemsCallbackData *item_data, gpointer data)
 static void
 mapi_msg_set_recipient_list (CamelMimeMessage *msg, MapiItem *item)
 {
-	g_return_if_fail (item->recipients != NULL);
-
 	GSList *l = NULL;
 	CamelInternetAddress *to_addr, *cc_addr, *bcc_addr;
 
+	g_return_if_fail (item->recipients != NULL);
+
 	to_addr = camel_internet_address_new ();
 	cc_addr = camel_internet_address_new ();
 	bcc_addr = camel_internet_address_new ();
 	
 	for (l = item->recipients; l; l=l->next) {
-		char *display_name, *name = NULL;
+		char *display_name;
+		const char *name = NULL;
 		uint32_t rcpt_type = MAPI_TO;
 		uint32_t *type = NULL; 
 		struct SRow aRow;
@@ -1080,7 +1355,7 @@ mapi_msg_set_recipient_list (CamelMimeMessage *msg, MapiItem *item)
 		type = (uint32_t *) find_SPropValue_data(&aRow, PR_RECIPIENT_TYPE);
 		
 		/*Fallbacks. Not good*/
-		display_name = name ? name : g_strdup (recip->email_id);
+		display_name = name ? g_strdup (name) : g_strdup (recip->email_id);
 		rcpt_type = (type ? *type : MAPI_TO);
 		
 		switch (rcpt_type) {
@@ -1098,7 +1373,7 @@ mapi_msg_set_recipient_list (CamelMimeMessage *msg, MapiItem *item)
 	}
 	
 	/*Add to message*/
-	camel_mime_message_set_recipients(msg, "To", to_addr);
+	/*Note : To field is added from PR_TRANSPORT_MESSAGE_HEADERS*/
 	camel_mime_message_set_recipients(msg, "Cc", cc_addr);
 	camel_mime_message_set_recipients(msg, "Bcc", bcc_addr);
 
@@ -1110,13 +1385,28 @@ static void
 mapi_populate_details_from_item (CamelFolder *folder, CamelMimeMessage *msg, MapiItem *item)
 {
 	char *temp_str = NULL;
+	const char *from_email;
 	time_t recieved_time;
 	CamelInternetAddress *addr = NULL;
 	CamelMapiStore *mapi_store = CAMEL_MAPI_STORE(folder->parent_store);
-
+	struct _camel_header_raw *headers = NULL;
+	gchar **header_list = NULL;
 	int offset = 0;
+	guint i;
 	time_t actual_time;
 
+	/* Setting headers from PR_TRANSPORT_MESSAGE_HEADERS */
+	
+	header_list = g_strsplit (item->header.transport_headers, "\n", i);
+	for (i = 0; header_list [i]; i++)
+		camel_header_raw_append_parse (&headers, header_list[i], -1);
+
+	while (headers->next) {
+		camel_medium_add_header (CAMEL_MEDIUM (msg), headers->name, headers->value);
+		headers = headers->next ;
+	}
+
+	/* Overwrite headers if we have specific properties available*/
 	temp_str = item->header.subject;
 	if(temp_str) 
 		camel_mime_message_set_subject (msg, temp_str);
@@ -1124,13 +1414,15 @@ mapi_populate_details_from_item (CamelFolder *folder, CamelMimeMessage *msg, Map
 	recieved_time = item->header.recieved_time;
 
 	actual_time = camel_header_decode_date (ctime(&recieved_time), &offset);
-	camel_mime_message_set_date (msg, actual_time, offset);
+	/* camel_mime_message_set_date (msg, actual_time, offset); */
 
 	if (item->header.from) {
 		if ((item->header.from_type != NULL) && !g_utf8_collate (item->header.from_type, "EX")) {
 			CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
-			item->header.from_email = exchange_mapi_util_ex_to_smtp (item->header.from_email);
+			from_email = exchange_mapi_util_ex_to_smtp (item->header.from_email);
 			CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+			g_free (item->header.from_email);
+			item->header.from_email = g_strdup (from_email);
 		}
 
 		item->header.from_email = item->header.from_email ? 
@@ -1158,16 +1450,27 @@ mapi_populate_msg_body_from_item (CamelMultipart *multipart, MapiItem *item, Exc
 	part = camel_mime_part_new ();
 	camel_mime_part_set_encoding(part, CAMEL_TRANSFER_ENCODING_8BIT);
 	
-	if (body) { 
+	if (body) {
+		gchar *buff = NULL;
+
 		if (item->is_cal)
-			camel_mime_part_set_content(part, (const char *) body->value->data, body->value->len, "text/calendar");
-		else {
+			type = "text/calendar";
+		else
 			type = (body->proptag == PR_BODY || body->proptag == PR_BODY_UNICODE) ? 
 				"text/plain" : "text/html";
 
-			/*NOTE : Last byte null mess up CRLF*. Probably needs a fix in e*fetch_items. */
-			camel_mime_part_set_content (part, (const char *) body->value->data, body->value->len - 1, type );
+		if (item->header.cpid) {
+			if (item->header.cpid == 65001)
+				buff = g_strdup_printf ("%s; charset=\"UTF-8\"", type);
+			else
+				buff = g_strdup_printf ("%s; charset=\"CP%d\"", type, item->header.cpid);
+			type = buff;
 		}
+
+		camel_mime_part_set_content (part, (const char *) body->value->data, body->value->len, type);
+
+		g_free (buff);
+		type = NULL;
 	} else
 		camel_mime_part_set_content(part, " ", strlen(" "), "text/html");
 
@@ -1178,23 +1481,20 @@ mapi_populate_msg_body_from_item (CamelMultipart *multipart, MapiItem *item, Exc
 
 static CamelMimeMessage *
 mapi_folder_item_to_msg( CamelFolder *folder,
-		MapiItem *item,
-		CamelException *ex )
+			 MapiItem *item,
+			 CamelException *ex )
 {
 	CamelMimeMessage *msg = NULL;
 	CamelMultipart *multipart = NULL;
 
 	GSList *attach_list = NULL;
-	int errno;
+	/* int errno; */
 	/* char *body = NULL; */
 	ExchangeMAPIStream *body = NULL;
 	GSList *body_part_list = NULL;
-	const char *uid = NULL;
 
 	attach_list = item->attachments;
-
 	msg = camel_mime_message_new ();
-
 	multipart = camel_multipart_new ();
 
 	/*FIXME : Using set of default. Fix it during mimewriter*/
@@ -1205,7 +1505,7 @@ mapi_folder_item_to_msg( CamelFolder *folder,
 
 	camel_multipart_set_boundary (multipart, NULL);
 
-	camel_mime_message_set_message_id (msg, uid);
+	/* Handle Multipart */
 	body_part_list = item->msg.body_parts;
 	while (body_part_list){
 	       body = body_part_list->data;
@@ -1213,11 +1513,21 @@ mapi_folder_item_to_msg( CamelFolder *folder,
 	       body_part_list = g_slist_next (body_part_list);
 	}
 
+	/* Threading */
+	if (item->header.message_id)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "Message-Id", item->header.message_id);
+
+	if (item->header.references)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "References", item->header.references);
+
+	if (item->header.in_reply_to)
+		camel_medium_add_header (CAMEL_MEDIUM (msg), "In-Reply-To", item->header.in_reply_to);
 
 	/*Set recipient details*/
 	mapi_msg_set_recipient_list (msg, item);
 	mapi_populate_details_from_item (folder, msg, item);
 
+	/** Attachment Handling*/
 	if (attach_list) {
 		GSList *al = attach_list;
 		for (al = attach_list; al != NULL; al = al->next) {
@@ -1284,8 +1594,6 @@ mapi_folder_get_message( CamelFolder *folder, const char *uid, CamelException *e
 	CamelMapiMessageInfo *mi = NULL;
 
 	CamelStream *stream, *cache_stream;
-	int errno;
-
 	mapi_id_t id_folder;
 	mapi_id_t id_message;
 	MapiItem *item = NULL;
@@ -1368,6 +1676,8 @@ mapi_folder_get_message( CamelFolder *folder, const char *uid, CamelException *e
 
 	msg = mapi_folder_item_to_msg (folder, item, ex);
 
+	exchange_mapi_util_free_recipient_list (&item->recipients);
+
 	g_free (item);
 
 	if (!msg) {
@@ -1437,11 +1747,6 @@ mapi_get_message_info(CamelFolder *folder, const char *uid)
 	}
 	msg_info = camel_message_info_new(folder->summary);
 	mi = (CamelMessageInfoBase *)msg_info ;
-	//TODO :
-/* 	oc_message_headers_init(&headers); */
-/* 	oc_thread_connect_lock(); */
-/* 	status = oc_message_headers_get_by_id(&headers, uid); */
-/* 	oc_thread_connect_unlock(); */
 
 	if (headers.subject) mi->subject = (char *)camel_pstring_strdup(headers.subject);
 	if (headers.from) mi->from = (char *)camel_pstring_strdup(headers.from);
@@ -1449,7 +1754,6 @@ mapi_get_message_info(CamelFolder *folder, const char *uid)
 	if (headers.cc) mi->cc = (char *)camel_pstring_strdup(headers.cc);
 	mi->flags = headers.flags;
 
-
 	mi->user_flags = NULL;
 	mi->user_tags = NULL;
 	mi->date_received = 0;
@@ -1627,6 +1931,66 @@ mapi_folder_rename (CamelFolder *folder, const char *new)
 	((CamelFolderClass *)parent_class)->rename(folder, new);
 }
 
+static gint
+mapi_cmp_uids (CamelFolder *folder, const char *uid1, const char *uid2)
+{
+	g_return_val_if_fail (uid1 != NULL, 0);
+	g_return_val_if_fail (uid2 != NULL, 0);
+
+	return strcmp (uid1, uid2);
+}
+
+static void
+mapi_append_message (CamelFolder *folder, CamelMimeMessage *message,
+		const CamelMessageInfo *info, gchar **appended_uid,
+		CamelException *ex)
+{
+	CamelMapiStore *mapi_store= CAMEL_MAPI_STORE(folder->parent_store);
+	CamelOfflineStore *offline = (CamelOfflineStore *) folder->parent_store;
+	CamelAddress *from = NULL, *recipients = NULL;
+	CamelStoreInfo *si;
+
+	MapiItem *item = NULL;
+	mapi_id_t fid = 0, mid = 0;
+	const gchar *folder_id;
+
+	/*Reject outbox / sent & trash*/
+	si = camel_store_summary_path ((CamelStoreSummary *)mapi_store->summary,
+				       folder->full_name);
+
+	if (si->flags & CAMEL_FOLDER_TYPE_TRASH ||
+	    si->flags & CAMEL_FOLDER_TYPE_OUTBOX) {
+		camel_exception_setv (ex, CAMEL_EXCEPTION_SYSTEM, 
+				      _("Cannot append message to folder '%s'"),
+				      folder->full_name);
+		return;
+	}
+
+	if (offline->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL) {
+		camel_exception_setv (ex, CAMEL_EXCEPTION_SYSTEM, 
+				 _("Offline. '%s'"));
+		return;
+	}
+
+	folder_id =  camel_mapi_store_folder_id_lookup (mapi_store,
+							folder->full_name);
+
+	exchange_mapi_util_mapi_id_from_string (folder_id, &fid);
+
+	/* Convert MIME to Item */
+	from = (CamelAddress *) camel_mime_message_get_from (message);
+	recipients = (CamelAddress *) camel_mime_message_get_recipients (message, "to");
+
+	item = camel_mapi_utils_mime_to_item (message, from, recipients, ex);
+
+	mid = exchange_mapi_create_item (olFolderOutbox, fid, NULL, NULL, 
+					 camel_mapi_utils_create_item_build_props, item,
+					 item->recipients, item->attachments,
+					 item->generic_streams, 0);
+
+	*appended_uid = exchange_mapi_util_mapi_ids_to_uid(fid, mid);
+}
+
 static void
 camel_mapi_folder_class_init (CamelMapiFolderClass *camel_mapi_folder_class)
 {
@@ -1639,10 +2003,11 @@ camel_mapi_folder_class_init (CamelMapiFolderClass *camel_mapi_folder_class)
 	camel_folder_class->get_message = mapi_folder_get_message;
  	camel_folder_class->rename = mapi_folder_rename;
 	camel_folder_class->search_by_expression = mapi_folder_search_by_expression;
+	camel_folder_class->cmp_uids = mapi_cmp_uids;
 /* 	camel_folder_class->get_message_info = mapi_get_message_info; */
 /* 	camel_folder_class->search_by_uids = mapi_folder_search_by_uids;  */
 	camel_folder_class->search_free = mapi_folder_search_free;
-/* 	camel_folder_class->append_message = mapi_append_message; */
+	camel_folder_class->append_message = mapi_append_message;
 	camel_folder_class->refresh_info = mapi_refresh_info;
 	camel_folder_class->sync = mapi_sync;
 	camel_folder_class->expunge = mapi_expunge;
@@ -1693,7 +2058,8 @@ camel_mapi_folder_get_type (void)
 }
 
 CamelFolder *
-camel_mapi_folder_new(CamelStore *store, const char *folder_name, const char *folder_dir, guint32 flags, CamelException *ex)
+camel_mapi_folder_new (CamelStore *store, const char *folder_name, const char *folder_dir,
+		      guint32 flags, CamelException *ex)
 {
 
 	CamelFolder	*folder = NULL;
@@ -1728,12 +2094,14 @@ camel_mapi_folder_new(CamelStore *store, const char *folder_name, const char *fo
 	}
 
 	/* set/load persistent state */
-	state_file = g_strdup_printf ("%s/cmeta", g_strdup_printf ("%s/%s",folder_dir, folder_name));
+	state_file = g_strdup_printf ("%s/%s/cmeta", folder_dir, folder_name);
 	camel_object_set(folder, NULL, CAMEL_OBJECT_STATE_FILE, state_file, NULL);
 	g_free(state_file);
 	camel_object_state_read(folder);
 
-	mapi_folder->cache = camel_data_cache_new (g_strdup_printf ("%s/%s",folder_dir, folder_name),0 ,ex);
+	state_file = g_strdup_printf ("%s/%s", folder_dir, folder_name);
+	mapi_folder->cache = camel_data_cache_new (state_file, 0, ex);
+	g_free (state_file);
 	if (!mapi_folder->cache) {
 		camel_object_unref (folder);
 		return NULL;
@@ -1747,10 +2115,8 @@ camel_mapi_folder_new(CamelStore *store, const char *folder_name, const char *fo
 /* 		return NULL; */
 /* 	} */
 
-	if (!strcmp (folder_name, "Mailbox")) {
-		if (camel_url_get_param (((CamelService *) store)->url, "filter"))
-			folder->folder_flags |= CAMEL_FOLDER_FILTER_RECENT;
-	}
+	if (camel_url_get_param (((CamelService *) store)->url, "filter"))
+		folder->folder_flags |= CAMEL_FOLDER_FILTER_RECENT;
 
 	mapi_folder->search = camel_folder_search_new ();
 	if (!mapi_folder->search) {
