@@ -70,6 +70,7 @@ mapi_profile_load (const char *profname, const char *password)
 	enum MAPISTATUS	retval = MAPI_E_SUCCESS;
 	struct mapi_session *session = NULL;
 	gchar *profpath = NULL;
+	gchar *default_profile_name = NULL;
 	const char *profile = NULL;
 	guint32 debug_log_level = 0;
 
@@ -100,13 +101,13 @@ mapi_profile_load (const char *profname, const char *password)
 	if (profname)
 		profile = profname;
 	else {
-		retval = GetDefaultProfile (&profile);
+		retval = GetDefaultProfile (&default_profile_name);
 		if (retval != MAPI_E_SUCCESS) {
 			mapi_errstr("GetDefaultProfile", GetLastError());
 			goto cleanup;
 		}
+		profile = default_profile_name;
 	}
-
 	g_print("\nLoading profile %s ", profile);
 
 	retval = MapiLogonEx(&session, profile, password);
@@ -116,8 +117,10 @@ mapi_profile_load (const char *profname, const char *password)
 	}
 
 cleanup:
-	if (retval != MAPI_E_SUCCESS && retval != MAPI_E_SESSION_LIMIT)
+	if (retval != MAPI_E_SUCCESS && retval != MAPI_E_SESSION_LIMIT &&
+	    retval != MAPI_E_LOGON_FAILED)
 		MAPIUninitialize ();
+
 	g_free (profpath);
 
 	d(g_print("\n%s: Leaving %s ", G_STRLOC, G_STRFUNC));
@@ -708,6 +711,87 @@ cleanup:
 	return status;
 }
 
+
+
+static ExchangeMAPIGALEntry * 
+mapidump_PAB_gal_entry (struct SRow *aRow)
+{
+	const char	*addrtype;
+	const char	*name;
+	const char	*email;
+	const char	*account;
+	ExchangeMAPIGALEntry *gal_entry;
+
+	addrtype = (const char *)find_SPropValue_data(aRow, PR_ADDRTYPE_UNICODE);
+	name = (const char *)find_SPropValue_data(aRow, PR_DISPLAY_NAME_UNICODE);
+	email = (const char *)find_SPropValue_data(aRow, PR_SMTP_ADDRESS_UNICODE);
+	account = (const char *)find_SPropValue_data(aRow, PR_ACCOUNT_UNICODE);
+
+	printf("[%s] %s:\n\tName: %-25s\n\tEmail: %-25s\n", 
+	       addrtype, account, name, email);
+
+	gal_entry = g_new0 (ExchangeMAPIGALEntry, 1);
+	gal_entry->name = g_strdup (name);
+	gal_entry->email = g_strdup (email);
+
+	return gal_entry;
+}
+
+gboolean
+exchange_mapi_util_get_gal (GPtrArray *contacts_array)
+{
+	struct SPropTagArray	*SPropTagArray;
+	struct SRowSet		*SRowSet;
+	enum MAPISTATUS		retval;
+	uint32_t		i;
+	uint32_t		count;
+	uint8_t			ulFlags;
+	TALLOC_CTX *mem_ctx;
+	
+	mem_ctx = talloc_init ("ExchangeMAPI_GetGAL");
+
+	SPropTagArray = set_SPropTagArray(mem_ctx, 0xc,
+					  PR_INSTANCE_KEY,
+					  PR_ENTRYID,
+					  PR_DISPLAY_NAME_UNICODE,
+					  PR_SMTP_ADDRESS_UNICODE,
+					  PR_DISPLAY_TYPE,
+					  PR_OBJECT_TYPE,
+					  PR_ADDRTYPE_UNICODE,
+					  PR_OFFICE_TELEPHONE_NUMBER_UNICODE,
+					  PR_OFFICE_LOCATION_UNICODE,
+					  PR_TITLE_UNICODE,
+					  PR_COMPANY_NAME_UNICODE,
+					  PR_ACCOUNT_UNICODE);
+
+	count = 0x7;
+	ulFlags = TABLE_START;
+	do {
+		count += 0x2;
+		retval = GetGALTable(global_mapi_session, SPropTagArray, &SRowSet, count, ulFlags);
+		if ((!SRowSet) || (!(SRowSet->aRow))) {
+			return false;
+		}
+		if (SRowSet->cRows) {
+			for (i = 0; i < SRowSet->cRows; i++) {
+				ExchangeMAPIGALEntry *gal_entry = g_new0 (ExchangeMAPIGALEntry, 1);
+				gal_entry = mapidump_PAB_gal_entry(&SRowSet->aRow[i]);
+				g_ptr_array_add(contacts_array, gal_entry);
+			}
+		}
+		ulFlags = TABLE_CUR;
+		MAPIFreeBuffer(SRowSet);
+	} while (SRowSet->cRows == count);
+	mapi_errstr("GetPABTable", GetLastError());
+
+	MAPIFreeBuffer(SPropTagArray);
+
+	return true;
+	
+}
+
+
+
 /* Returns TRUE if all recipients were read succcesfully, else returns FALSE */
 static gboolean
 exchange_mapi_util_get_recipients (mapi_object_t *obj_message, GSList **recip_list)
@@ -1156,6 +1240,8 @@ exchange_mapi_connection_fetch_items   (mapi_id_t fid,
 			if (options & MAPI_OPTIONS_FETCH_RECIPIENTS) 
 				exchange_mapi_util_get_recipients (&obj_message, &recip_list);
 
+//			exchange_mapi_util_get_gal (contacts_array);
+
 			/* get the main body stream no matter what */
 			if (options & MAPI_OPTIONS_FETCH_BODY_STREAM)
 				exchange_mapi_util_read_body_stream (&obj_message, &stream_list, 
@@ -2459,12 +2545,13 @@ get_child_folders(TALLOC_CTX *mem_ctx, ExchangeMAPIFolderCategory folder_hier, m
 		goto cleanup;
 	}
 
-	SPropTagArray = set_SPropTagArray(mem_ctx, 0x6,
+	SPropTagArray = set_SPropTagArray(mem_ctx, 0x7,
 					  PR_FID,
 					  PR_CONTAINER_CLASS,
 					  PR_DISPLAY_NAME,
 					  PR_CONTENT_UNREAD,
 					  PR_CONTENT_COUNT,
+					  PR_MESSAGE_SIZE,
 					  PR_FOLDER_CHILD_COUNT);
 
 	retval = SetColumns(&obj_table, SPropTagArray);
@@ -2492,17 +2579,20 @@ get_child_folders(TALLOC_CTX *mem_ctx, ExchangeMAPIFolderCategory folder_hier, m
 		const uint32_t *unread = (const uint32_t *)find_SPropValue_data(&rowset.aRow[i], PR_CONTENT_UNREAD);
 		const uint32_t *total = (const uint32_t *)find_SPropValue_data(&rowset.aRow[i], PR_CONTENT_COUNT);
 		const uint32_t *child = (const uint32_t *)find_SPropValue_data(&rowset.aRow[i], PR_FOLDER_CHILD_COUNT);
+		const uint32_t *folder_size = (const uint32_t *)find_SPropValue_data(&rowset.aRow[i], PR_MESSAGE_SIZE);
 
 		if (!class)
 			class = IPF_NOTE;
 
 		newname = utf8tolinux (name);
-		g_print("\n|---+ %-15s : (Container class: %s %016" G_GINT64_MODIFIER "X) UnRead : %d Total : %d ", 
-			newname, class, *fid, unread ? *unread : 0, total ? *total : 0);
+		g_print("\n|---+ %-15s : (Container class: %s %016" G_GINT64_MODIFIER "X) UnRead : %d Total : %d size : %d", 
+			newname, class, *fid, unread ? *unread : 0, total ? *total : 0, folder_size ? *folder_size : 0);
 
 		folder = exchange_mapi_folder_new (newname, class, folder_hier, *fid, folder_id,
 						   child ? *child : 0, unread ? *unread : 0, total ? *total : 0);
 
+		folder->size = folder_size ? *folder_size : 0;
+
 		*mapi_folders = g_slist_prepend (*mapi_folders, folder);
 
 		if (child && *child && (depth != 0))
@@ -2669,6 +2759,7 @@ exchange_mapi_get_folders_list (GSList **mapi_folders)
 	char 			*utf8_mailbox_name = NULL;
 	const char 		*mailbox_owner_name = NULL;
 	const char 		*mailbox_user_name = NULL;
+	const uint32_t          *mailbox_size = NULL;
 
 	d(g_print("\n%s: Entering %s ", G_STRLOC, G_STRFUNC));
 
@@ -2685,9 +2776,10 @@ exchange_mapi_get_folders_list (GSList **mapi_folders)
 	}
 
 	/* Build the array of Mailbox properties we want to fetch */
-	SPropTagArray = set_SPropTagArray(mem_ctx, 0x3,
+	SPropTagArray = set_SPropTagArray(mem_ctx, 0x4,
 					  PR_DISPLAY_NAME,
 					  PR_MAILBOX_OWNER_NAME,
+					  PR_MESSAGE_SIZE,
 					  PR_USER_NAME);
 
 	lpProps = talloc_zero(mem_ctx, struct SPropValue);
@@ -2708,6 +2800,7 @@ exchange_mapi_get_folders_list (GSList **mapi_folders)
 	mailbox_name = (const char *) find_SPropValue_data(&aRow, PR_DISPLAY_NAME);
 	mailbox_owner_name = (const char *) find_SPropValue_data(&aRow, PR_MAILBOX_OWNER_NAME);
 	mailbox_user_name = (const char *) find_SPropValue_data(&aRow, PR_USER_NAME);
+	mailbox_size = (const uint32_t *)find_SPropValue_data (&aRow, PR_MESSAGE_SIZE);
 
 	/* Prepare the directory listing */
 	retval = GetDefaultFolder(&obj_store, &mailbox_id, olFolderTopInformationStore);
@@ -2723,6 +2816,7 @@ exchange_mapi_get_folders_list (GSList **mapi_folders)
 					   MAPI_PERSONAL_FOLDER, mailbox_id, 0, 0, 0 ,0); 
 	folder->is_default = true;
 	folder->default_type = olFolderTopInformationStore; /*Is this correct ?*/
+	folder->size = *mailbox_size;
 
 	*mapi_folders = g_slist_prepend (*mapi_folders, folder);
 
@@ -2880,8 +2974,11 @@ manage_mapi_error (const char *context, uint32_t error_id, char **error_msg)
 	}
 }
 
+
 gboolean
-exchange_mapi_create_profile (const char *username, const char *password, const char *domain, const char *server, char **error_msg)
+exchange_mapi_create_profile (const char *username, const char *password, const char *domain,
+			      const char *server, char **error_msg,
+			      mapi_profile_callback_t callback, gpointer data)
 {
 	enum MAPISTATUS	retval;
 	gboolean result = FALSE; 
@@ -2956,13 +3053,16 @@ exchange_mapi_create_profile (const char *username, const char *password, const
 			manage_mapi_error ("DeleteProfile", GetLastError(), error_msg);
 		goto cleanup; 
 	}
-	d(g_print("succeeded \n"));
+	d(g_print("MapiLogonProvider : succeeded \n"));
 
-	retval = ProcessNetworkProfile(session, username, NULL, NULL); 
+	retval = ProcessNetworkProfile(session, username, callback, NULL); 
 	if (retval != MAPI_E_SUCCESS) {
 		manage_mapi_error ("ProcessNetworkProfile", GetLastError(), error_msg);
-		goto cleanup; 
+		g_debug ("Deleting profile %s ", profname); 
+		DeleteProfile(profname); 
+		goto exit; 
 	}
+	d(g_print("ProcessNetworkProfile : succeeded \n"));
 
 	/* Set it as the default profile. Is this needed? */
 	retval = SetDefaultProfile(profname); 
@@ -2978,12 +3078,14 @@ exchange_mapi_create_profile (const char *username, const char *password, const
 	if (exchange_mapi_connection_new (profname, password)) {
 		result = TRUE;
 		exchange_mapi_peek_folder_list ();
-	}
+	} else 
+		goto exit;
 
 cleanup: 
 	if (!result)
 		MAPIUninitialize ();
 
+exit:
 	g_free (profname);
 	g_free (profpath);
 
